#!groovy
/*
  README
  ======

  Parameterised variables are used to define the build configuration.
  Properties used are string or choice, for a choice, the first entry is default.

*/
properties([[
  $class: "ParametersDefinitionProperty",
  parameterDefinitions: [
  string(name: "label", defaultValue: "default", description: "Label for Slack fallback"),
  string(name: "repo_host", defaultValue: "bitbucket.org", description: "Domain for repository"),
  string(name: "repo_path", defaultValue: "gyrokinetics/gs2", description: "Relative path to repository"),
  choice(name: "repo_method", choices: "ssh\nhttps", description: "Protocol used to access repository"),
  choice(name: "opt", choices: "intermediate\ndebug\nfast\nwarnings\ncoverage", description: "Build type"),
  choice(name: "comms", choices: "serial\nmpi", description: "Communications paradigm"),
  string(name: "testexec", defaultValue: "mpirun -np 2", description: "Test command"),
  string(name: "modules", defaultValue: "gcc/4.8.5 fftw python/3.6.0 openmpi/1.10.0", description: "Module environments to load"),
  string(name: "compiler", defaultValue: "gcc-4.8.5", description: "compiler-version, for consitent path names"),
  string(name: "GK_SYSTEM", defaultValue: "jenkins-4.8.5", description: "Value of GK_SYSTEM, which determines the local Makefile used"),
  string(name: "fft", defaultValue: "fftw3", description: "Select FFT to use"),
  string(name: "baseURL", defaultValue: "https://gyrokinetics.slack.com/services/hooks/", description: "Base URL for Slack webhook")
]]])

/*
 End README
*/

// Declare variables
String slackURL
String startMessage
String channel
String payload
String command
String colourCode
String result
String resultMessage
String fallback
String regressionDetails = ""
String netcdff_path
int numErrors = 0
int testFailures = 0
int testSkipped = 0
boolean buildSuccess = true
String PRname
String PRnameLink
String PRnameNoLink
String jobDescription
String jobDescriptionNoLink
String outcome

import groovy.json.JsonBuilder // For json construction.
  
// Do all the work on a node
node ('sl7') {

// PR name for display in Jenkins
PRname = currentBuild.description
// PR name for display in Slack
PRnameNoLink = "#${env.pullRequestId} ${env.pullRequestTitle}"
PRnameLink = "<https://${env.repo_host}/${env.repo_path}/pull-requests/${env.pullRequestId}|"+PRnameNoLink+">"

// Description of build and PR used in Slack messages
jobDescription = "Anvil build <${env.JOB_URL}/${env.BUILD_NUMBER}|#${env.BUILD_NUMBER}> for suite <${env.JOB_URL}|${env.JOB_NAME}> and PR "+PRnameLink
// Text-only version of jobDescription
jobDescriptionNoLink = "Anvil build #${env.BUILD_NUMBER} for suite ${env.JOB_NAME} and PR "+PRnameNoLink

// Set up for Slack API
withCredentials([string(credentialsId: 'slack-id', variable: 'token')]) { 
	slackURL = "${params.baseURL}jenkins-ci?token=${token}"
}

startMessage = "Anvil build #${env.BUILD_NUMBER} for ${env.JOB_NAME}. (<${env.JOB_URL}|View>)"

// This fallback message shows in a Slack notification

fallback = "gs2 ${params.label} \"${params.opt}\" ${params.comms} build started for branch ${params.branch}"

channel = "#jenkins"
colourCode = "#439FE0" // Jenkins blue

// Construct the json payload
def mrkdwn = [ "pretext", "text", "fields" ]  // Required for Slack API
def attachment = [pretext: fallback,
                  text: startMessage,
		  fallback: fallback,
		  color: colourCode,
		  mrkdwn_in: mrkdwn]

def attachments = [ attachment ]
def load = [ channel: channel, attachments: attachments ]

try {

  def json = new JsonBuilder(load) // This may fail if the script isn't whitelisted
  payload = json.toString()
  json = null // Need to nullify the Json object as it is not "serializable" by Java

}

catch (jsonError) {

  echo "Error constructing json for Slack message - likely a script whitelist issue"



  // Construct the json as a string - much easier to get wrong, much harder to debug

  payload = "{\"channel\":\"#jenkins\",\"attachments\":[{\"pretext\":\"${fallback}\",\"text\":\"${startMessage}\",\"fallback\":\"${fallback}\",\"color\":\"#439FE0\",\"mrkdwn_in\":[\"pretext\",\"text\",\"fields\"]}]}"

}

// Construct the command as a string so that variables are expanded properly
command = "set +x; curl -X POST -H 'Content-Type: application/json' -d '${payload}' ${slackURL}"

// Set up netcdf paths
String NETCDF_VERSION="4.6.1"
String NETCDFF_VERSION="4.4.4"
String MY_NETCDF_DIR="${WORKSPACE}/software/${params.compiler}/netcdf/${NETCDF_VERSION}"
String MY_NETCDFF_DIR="${WORKSPACE}/software/${params.compiler}/netcdf-fortran/${NETCDFF_VERSION}"
String MY_NETCDF_INC="${MY_NETCDF_DIR}/include"
String MY_NETCDFF_INC="${MY_NETCDFF_DIR}/include"
String MY_NETCDF_LIB="${MY_NETCDF_DIR}/lib"
String MY_NETCDFF_LIB="${MY_NETCDFF_DIR}/lib"
String libpath="${MY_NETCDF_LIB}:${MY_NETCDFF_LIB}:/opt/software/gcc-7.2.0/openmpi/1.10.0/lib:/opt/software/gcc-7.2.0/fftw/3.3.5/lib:/opt/software/gcc/7.2.0/lib64:/opt/software/gcc/7.2.0/lib"


	stage('Checkout') {

        	currentBuild.description = "Checking out from repository"

        	// Old directory is kept
        	sh "rm -rf gs2"

		checkout scm

        	sh '''
			git config --global user.email "joseph.parker@stfc.ac.uk"
			git config --global user.name "Joseph Parker"
           		git submodule update --init --recursive
		'''
    	}

	stage('Merge') {

        	currentBuild.description = PRname + " - Merging PR"
		
		try{
			sh '.jenkins/merge.sh'
		} catch(gitError) {
        		currentBuild.description = PRname + " - PR merge failed"
			currentBuild.result = 'FAILURE'
	
			runDuration=currentBuild.durationString[0..-14]
        		endMessage = ""
			colourCode = '#FF0000' // failure red
    			resultMessage = "Anvil build <${env.JOB_URL}/${env.BUILD_NUMBER}|#${env.BUILD_NUMBER}> for <${env.JOB_URL}|${env.JOB_NAME}> failed in git merge step after ${runDuration}"
    			fallback = "Anvil build #${env.BUILD_NUMBER} for ${env.JOB_NAME} failed in git merge step after ${runDuration}"

  			attachment = [text: resultMessage,
				pretext: endMessage,
				fallback: fallback,
				color: colourCode,
				mrkdwn_in: mrkdwn]
  			attachments = [ attachment ]
  			load = [ channel: channel, attachments: attachments ]
  			try {
    				endJson = new JsonBuilder(load)
    				payload = endJson.toString()
 		   		endJson = null // Need to nullify the Json object as it is not "serializable" by Java
 		 	} catch (jsonError) {
    				echo "Error contructing final Slack message - likely a library permissions issue"
		    		payload = "{\"channel\":\"#jenkins\",\"attachments\":[{\"Job Ended\",\"text\":\"${resultMessage}\",\"fallback\":\"Job Ended\",\"color\":\"${colourCode}\",\"mrkdwn_in\":[\"pretext\",\"text\",\"fields\"]}]}"
  			}
  			command = "set +x; curl -X POST -H 'Content-Type: application/json' -d '${payload}' ${slackURL}"
			try {
				sh command
			} catch (Exception slackError) {
				echo "Error posting to Slack"
			}
			throw gitError
		}
    	}
	
  stage('Install netcdff') {

        currentBuild.description = "Installing netcdf/netcdff, if necessary"

        sh """

        export TEST_FILE_NETCDF=${MY_NETCDF_DIR}/include/netcdf.h
        export TEST_FILE_NETCDFF=${MY_NETCDFF_DIR}/include/netcdf.mod

        # true if either file missing
        if [[ -f \$TEST_FILE_NETCDF && -f \$TEST_FILE_NETCDFF ]]
        then 
              echo "netcdf + nectdff found"
        else
              module load ${params.modules}

              rm -rf ${MY_NETCDF_DIR}
              rm -rf ${MY_NETCDFF_DIR}

              mkdir -p ${MY_NETCDF_DIR}
              wget https://github.com/Unidata/netcdf-c/archive/v${NETCDF_VERSION}.tar.gz
              tar xvf v${NETCDF_VERSION}.tar.gz
              cd netcdf-c-${NETCDF_VERSION}
              ./configure --prefix=${MY_NETCDF_DIR}
              make
              make install
              export LD_LIBRARY_PATH=\$LD_LIBRARY_PATH:${MY_NETCDF_LIB}
              export CPPFLAGS=-I${MY_NETCDF_INC}
              export LDFLAGS=-L${MY_NETCDF_LIB}
              cd ..

              mkdir -p ${MY_NETCDFF_DIR}
              wget https://github.com/Unidata/netcdf-fortran/archive/v${NETCDFF_VERSION}.tar.gz
              tar xvf v${NETCDFF_VERSION}.tar.gz
              cd netcdf-fortran-${NETCDFF_VERSION}
              ./configure --prefix=${MY_NETCDFF_DIR}
              make
              make install
              export LD_LIBRARY_PATH=\$LD_LIBRARY_PATH:${MY_NETCDFF_LIB}

	      ${MY_NETCDFF_DIR}/bin/nf-config --all
              echo \$LD_LIBRARY_PATH
	fi
        """
  }

	stage('Build') {

        	currentBuild.description = "Building"
		
		try{
			sh """
				module load ${params.modules}
				export GK_SYSTEM=${params.GK_SYSTEM}
    				export LD_LIBRARY_PATH=\$LD_LIBRARY_PATH:${libpath}
                              	export MAKEFLAGS=-IMakefiles
                              	export TESTEXEC='mpirun -np 2'

				make depend
				make test_make
				make -j
			"""
		} catch(buildError) {
        		currentBuild.description = PRname + " - build failed"
			currentBuild.result = 'FAILURE'
			throw buildError
		}

    	}

	stage('Test') {

        	currentBuild.description = "Testing"
		try{
			sh """
				export TESTEXEC="${params.testexec}"
				module load ${params.modules}
				export GK_SYSTEM=${params.GK_SYSTEM}
    				export LD_LIBRARY_PATH=\$LD_LIBRARY_PATH:${libpath}
                              	export MAKEFLAGS=-IMakefiles
				source .jenkins/set_python_env.sh
				make tests TEST_LEVEL=2
				cd ../
				rm -rf gs2
			"""
		} catch(testError) {
        		currentBuild.description = PRname + " - tests failed"
			currentBuild.result = 'FAILURE'
			throw testError
		}
    	}

    stage('Finishing') {

        currentBuild.description = "Parsing test results"
	testFailures = sh ( script: """ grep 'failures="1"' test-results/*.xml | wc -l """, returnStdout: true ) as Integer

        currentBuild.description = "Sending notification to slack"
	
	// "currentBuild.durationString" unhelpfully returns the time followed by "and counting" because we're still in the pipeline. Bah.
	runDuration=currentBuild.durationString[0..-14]
        endMessage = ""
	colourCode = '#FF0000' // failure red
    	resultMessage = "Anvil build <${env.JOB_URL}/${env.BUILD_NUMBER}|#${env.BUILD_NUMBER}> for <${env.JOB_URL}|${env.JOB_NAME}> finished in mysterious circumstances."
    	fallback = "Anvil build #${env.BUILD_NUMBER} for ${env.JOB_NAME} finished in mysterious circumstances."

	result = currentBuild.result ?: 'SUCCESS'

  	if (result == 'ABORTED') {
    		colourCode = '#808080' // anonymous grey
    		currentBuild.description = "Build aborted"
    		resultMessage = "Anvil build <${env.JOB_URL}/${env.BUILD_NUMBER}|#${env.BUILD_NUMBER}> for <${env.JOB_URL}|${env.JOB_NAME}> was aborted."
        	fallback = "Anvil build #${env.BUILD_NUMBER} for ${env.JOB_NAME} was aborted." // same message but without hyperlinks, for pop up notifications
		currentBuild.result = 'ABORTED'
  	}
  	else if (testFailures > 0) {
    		colourCode = '#FF0000'
    		currentBuild.description = "Build fails"
		outcome = " finished with ${testFailures} failures in ${runDuration}."
    		resultMessage = jobDescription+outcome
        	fallback = jobDescriptionNoLink+outcome // same message but without hyperlinks, for pop up notifications
        	endMessage = '' // No title
		currentBuild.result = 'FAILURE'
	}
  	else if (testFailures == 0) {
    		colourCode = '#32CD32' // lovely successful green
    		currentBuild.description = "Stable build"
		outcome = " finished successfully in ${runDuration}."
    		resultMessage = jobDescription+outcome
        	fallback = jobDescriptionNoLink+outcome // same message but without hyperlinks, for pop up notifications
        	endMessage = '' // No title
	}
  	else {
		println "Something is wrong with our counting"
  	}

  	attachment = [text: resultMessage,
		pretext: endMessage,
		fallback: fallback,
		color: colourCode,
		mrkdwn_in: mrkdwn]
  	attachments = [ attachment ]
  	load = [ channel: channel, attachments: attachments ]
  	try {
    		endJson = new JsonBuilder(load)
    		payload = endJson.toString()
    		endJson = null // Need to nullify the Json object as it is not "serializable" by Java
  	} catch (jsonError) {
    		echo "Error contructing final Slack message - likely a library permissions issue"
    		payload = "{\"channel\":\"#jenkins\",\"attachments\":[{\"Job Ended\",\"text\":\"${resultMessage}\",\"fallback\":\"Job Ended\",\"color\":\"${colourCode}\",\"mrkdwn_in\":[\"pretext\",\"text\",\"fields\"]}]}"
  	}
  	command = "set +x; curl -X POST -H 'Content-Type: application/json' -d '${payload}' ${slackURL}"
	try {
		sh command
	} catch (slackError) {
		echo "Error posting to Slack"
	}

	// Now everything is finished, set the name back to a useful hyperlink:
	currentBuild.description = PRname

    }
}

